@@ -26,14 +26,13 @@ use strict;
 #
 # @brief A parser for the disk_config files within FAI.
 #
-# $Id$
-#
 # @author Christian Kern, Michael Tautschnig, Sam Vilain, Andreas Schludei
 # @date Sun Jul 23 16:09:36 CEST 2006
 #
 ################################################################################
 
 use Parse::RecDescent;
+use Storable qw(dclone);
 
 package FAI;
 
@@ -48,7 +47,8 @@ $FAI::device = "";
 
 ################################################################################
 #
-# @brief Test, whether @ref $cmd is available on the system using $PATH
+# @brief Test, whether @ref $cmd is available and executable on the system using
+#        $PATH
 #
 # @param $cmd Command that is to be found in $PATH
 #
@@ -60,6 +60,12 @@ sub in_path {
   # initialize the parameter
   my ($cmd) = @_;
 
+  # ignored in syntax-check mode
+  return 1 if ($FAI::check_only);
+
+  # check full path names first
+  ($cmd =~ /^\//) and return (-x "$cmd");
+
   # split $PATH into its components, search all of its components
   # and test for $cmd being executable
   (-x "$_/$cmd") and return 1 foreach (split (":", $ENV{PATH}));
@@ -80,10 +86,13 @@ sub in_path {
 ################################################################################
 sub resolve_disk_shortname {
   my ($disk) = @_;
+  $disk =~ s/^disk//;
+
+  $disk = "sdx" . chr(ord('a') + $disk - 1)
+    if ($FAI::check_only && $disk =~ /^\d+$/);
 
   # test $disk for being numeric
   if ($disk =~ /^\d+$/) {
-
     # $disk-1 must be a valid index in the map of all disks in the system
     (scalar(@FAI::disks) >= $disk)
       or die "this system does not have a physical disk $disk\n";
@@ -122,12 +131,15 @@ sub init_disk_config {
 
   $disk = &FAI::resolve_disk_shortname($disk);
 
+  &FAI::in_path("losetup") or die "losetup not found in PATH\n"
+    if ((&FAI::loopback_dev($disk))[0]);
+
   # prepend PHY_
   $FAI::device = "PHY_$disk";
 
   # test, whether this is the first disk_config stanza to configure $disk
   defined ($FAI::configs{$FAI::device})
-    and die "Duplicate configuration for disk $FAI::disks[ $1-1 ]\n";
+    and die "Duplicate configuration for disk $disk\n";
 
   # Initialise the entry in $FAI::configs
   $FAI::configs{$FAI::device} = {
@@ -136,15 +148,22 @@ sub init_disk_config {
     bootable   => -1,
     fstabkey   => "device",
     preserveparts => 0,
-    partitions => {}
+    partitions => {},
+    opts_all   => {}
   };
+
+  # Init device tree object
+  $FAI::dev_children{$disk} = ();
+
+  return 1;
 }
 
 ################################################################################
 #
 # @brief Initialise the entry of a partition in @ref $FAI::configs
 #
-# @param $type The type of the partition. It must be either primary or logical.
+# @param $type The type of the partition. It must be either primary or logical
+# or raw.
 #
 ################################################################################
 sub init_part_config {
@@ -152,22 +171,36 @@ sub init_part_config {
   # the type of the partition to be created
   my ($type) = @_;
 
-  # type must either be primary or logical, nothing else may be accepted by the
-  # parser
-  ($type eq "primary" || $type eq "logical") or 
+  # type must either be primary or logical or raw, nothing else may be accepted
+  # by the parser
+  ($type eq "primary" || $type eq "logical" || $type eq "raw") or
     &FAI::internal_error("invalid type $type");
 
   # check that a physical device is being configured; logical partitions are
   # only supported on msdos disk labels.
-  ($FAI::device =~ /^PHY_/ && ($type ne "logical"
-      || $FAI::configs{$FAI::device}{disklabel} eq "msdos")) or 
+  ($FAI::device =~ /^PHY_(.+)$/ && ($type ne "logical"
+      || $FAI::configs{$FAI::device}{disklabel} eq "msdos")) or
     die "Syntax error: invalid partition type";
 
+  # the disk
+  my $disk = $1;
+
   # the index of the new partition
   my $part_number = 0;
 
+  # defaults from options
+  my $preserve_default =
+    defined($FAI::configs{$FAI::device}{opts_all}{preserve}) ? 1 :
+      (defined($FAI::configs{$FAI::device}{opts_all}{preserve_lazy}) ? 2 : 0);
+  my $always_format_default =
+    defined($FAI::configs{$FAI::device}{opts_all}{always_format}) ? 1 : 0;
+  my $resize_default =
+    defined($FAI::configs{$FAI::device}{opts_all}{resize}) ? 1 : 0;
+
   # create a primary partition
   if ($type eq "primary") {
+    (defined($FAI::configs{$FAI::device}{partitions}{0})) and
+      die "You cannot use raw-disk together with primary/logical partitions\n";
 
     # find all previously defined primary partitions
     foreach my $part_id (&numsort(keys %{ $FAI::configs{$FAI::device}{partitions} })) {
@@ -187,10 +220,17 @@ sub init_part_config {
     $part_number++;
 
     # msdos disk labels don't allow for more than 4 primary partitions
-    ($part_number < 5 || $FAI::configs{$FAI::device}{virtual} || 
+    ($part_number < 5 || $FAI::configs{$FAI::device}{virtual} ||
       $FAI::configs{$FAI::device}{disklabel} ne "msdos")
       or die "$part_number are too many primary partitions\n";
+  } elsif ($type eq "raw") {
+    (0 == scalar(keys %{ $FAI::configs{$FAI::device}{partitions} })) or
+      die "You cannot use raw-disk together with primary/logical partitions\n";
+    # special-case hack: part number 0 is invalid otherwise
+    $part_number = 0;
   } else {
+    (defined($FAI::configs{$FAI::device}{partitions}{0})) and
+      die "You cannot use raw-disk together with primary/logical partitions\n";
 
     # no further checks for the disk label being msdos have to be performed in
     # this branch, it has been ensured above
@@ -246,10 +286,6 @@ sub init_part_config {
           size => {}
         };
 
-      # as we can't compute the index from the reference, we need to store the
-      # $part_number explicitly
-      (\%FAI::configs)->{$FAI::device}->{partitions}->{$extended}->{number} = $extended;
-
       my $part_size =
         (\%FAI::configs)->{$FAI::device}->{partitions}->{$extended}->{size};
 
@@ -264,8 +300,11 @@ sub init_part_config {
       defined ($part_size->{always_format})
         or $part_size->{always_format} = 0;
 
-      # add the resize = 0 flag, if it doesn't exist already
-      defined ($part_size->{resize}) or $part_size->{resize} = 0;
+      # add the resize = default flag, if it doesn't exist already
+      defined ($part_size->{resize}) or $part_size->{resize} = $resize_default;
+
+      # add entry to device tree
+      push @{ $FAI::dev_children{$disk} }, &FAI::make_device_name($disk, $extended);
     }
   }
 
@@ -279,25 +318,25 @@ sub init_part_config {
   # the reference is used by all further processing of this config line
   $FAI::partition_pointer =
     (\%FAI::configs)->{$FAI::device}->{partitions}->{$part_number};
-
-  # as we can't compute the index from the reference, we need to store the
-  # $part_number explicitly
-  $FAI::partition_pointer->{number} = $part_number;
+  $FAI::partition_pointer_dev_name = &FAI::make_device_name($disk, $part_number);
 
   # the partition is not an extended one
   $FAI::partition_pointer->{size}->{extended} = 0;
 
   # add the preserve = 0 flag, if it doesn't exist already
   defined ($FAI::partition_pointer->{size}->{preserve})
-    or $FAI::partition_pointer->{size}->{preserve} = 0;
+    or $FAI::partition_pointer->{size}->{preserve} = $preserve_default;
 
   # add the always_format = 0 flag, if it doesn't exist already
   defined ($FAI::partition_pointer->{size}->{always_format})
-    or $FAI::partition_pointer->{size}->{always_format} = 0;
+    or $FAI::partition_pointer->{size}->{always_format} = $always_format_default;
 
   # add the resize = 0 flag, if it doesn't exist already
   defined ($FAI::partition_pointer->{size}->{resize})
-    or $FAI::partition_pointer->{size}->{resize} = 0;
+    or $FAI::partition_pointer->{size}->{resize} = $resize_default;
+
+  # add entry to device tree
+  push @{ $FAI::dev_children{$disk} }, $FAI::partition_pointer_dev_name;
 }
 
 ################################################################################
@@ -326,24 +365,58 @@ sub convert_unit
       return $val * $totalmem;
   }
 
-  ## don't warn for now, G/GiB/GB are all treated the same way
-  ## ($val =~ /([kKMGTP])\s*$/) and
-  ##   warn "Using $1 as size modifier is deprecated, please use $1iB or $1B
-  ##   instead; in future releases these will be treated as different modifiers\n";
-
   # % is returned as is
   if ($val =~ /^(\d+(\.\d+)?)%\s*$/) { 1; }
   elsif ($val =~ /^(\d+(\.\d+)?)B\s*$/) { $val = $1 * (1 / 1024) * (1 / 1024); }
-  elsif ($val =~ /^(\d+(\.\d+)?)[kK](i)?(B)?\s*$/) { $val = $1 * (1 / 1024); }
-  elsif ($val =~ /^(\d+(\.\d+)?)M(i)?(B)?\s*$/) { $val = $1; }
-  elsif ($val =~ /^(\d+(\.\d+)?)G(i)?(B)?\s*$/) { $val = $1 * 1024; }
-  elsif ($val =~ /^(\d+(\.\d+)?)T(i)?(B)?\s*$/) { $val = $1 * (1024 * 1024); }
-  elsif ($val =~ /^(\d+(\.\d+)?)P(i)?(B)?\s*$/) { $val = $1 * (1024 * 1024 * 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)[kK](iB)?\s*$/) { $val = $1 * (1 / 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)[kK]B\s*$/) { $val = $1 * (1000 / 1024) * (1 / 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)M(iB)?\s*$/) { $val = $1; }
+  elsif ($val =~ /^(\d+(\.\d+)?)MB\s*$/) { $val = $1 * (1000 / 1024) * (1000 / 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)G(iB)?\s*$/) { $val = $1 * 1024; }
+  elsif ($val =~ /^(\d+(\.\d+)?)GB\s*$/) { $val = $1 * 1000 * (1000 / 1024) * (1000 / 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)T(iB)?\s*$/) { $val = $1 * (1024 * 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)TB\s*$/) { $val = $1 * 1000 * 1000 * (1000 / 1024) * (1000 / 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)P(iB)?\s*$/) { $val = $1 * (1024 * 1024 * 1024); }
+  elsif ($val =~ /^(\d+(\.\d+)?)PB\s*$/) { $val = $1 * 1000 * 1000 * 1000 * (1000 / 1024) * (1000 / 1024); }
   else { &FAI::internal_error("convert_unit $val"); }
 
   return $val;
 }
 
+################################################################################
+#
+# @brief Fill the "size" key of a partition or volume entry
+#
+# @param $range Actual size
+# @param $options Additional options such as preserve or resize
+#
+################################################################################
+sub set_volume_size
+{
+  my ($range, $options) = @_;
+
+  # convert the units, if necessary
+  my ($min, $max) = split (/-/, $range);
+  $min .= "MiB" if ($min =~ /\d\s*$/);
+  $min   = &FAI::convert_unit($min);
+  $max .= "MiB" if ($max =~ /\d\s*$/);
+  $max   = &FAI::convert_unit($max);
+  # enter the range into the hash
+  $FAI::partition_pointer->{size}->{range} = "$min-$max";
+  # set the resize or preserve flag, if required
+  if (defined ($options)) {
+    $FAI::configs{$FAI::device}{preserveparts} = 1;
+    $FAI::partition_pointer->{size}->{resize} = 1 if ($options =~ /^:resize/);
+    $FAI::partition_pointer->{size}->{preserve} = 1 if ($options =~ /^:preserve_always/);
+    $FAI::partition_pointer->{size}->{preserve} = 1
+    if ($FAI::reinstall && $options =~ /^:preserve_reinstall/);
+    if ($options =~ /^:preserve_lazy/) {
+      $FAI::configs{$FAI::device}{preserveparts} = 2;
+      $FAI::partition_pointer->{size}->{preserve} = 2;
+    }
+  }
+}
+
 # have RecDescent do proper error reporting
 $::RD_HINT = 1;
 
@@ -362,6 +435,7 @@ $FAI::Parser = Parse::RecDescent->new(
     line: <skip: qr/[ \t]*/> "\\n"
         | <skip: qr/[ \t]*/> comment "\\n"
         | <skip: qr/[ \t]*/> config "\\n"
+        | <error>
 
     comment: /^\s*#.*/
 
@@ -374,6 +448,7 @@ $FAI::Parser = Parse::RecDescent->new(
           &FAI::in_path("mdadm") or die "mdadm not found in PATH\n";
           $FAI::device = "RAID";
           $FAI::configs{$FAI::device}{fstabkey} = "device";
+          $FAI::configs{$FAI::device}{opts_all} = {};
         }
         raid_option(s?)
         | 'cryptsetup'
@@ -394,6 +469,7 @@ $FAI::Parser = Parse::RecDescent->new(
           # being configured
           $FAI::device = "VG_";
           $FAI::configs{"VG_--ANY--"}{fstabkey} = "device";
+          $FAI::configs{"VG_--ANY--"}{opts_all} = {};
         }
         lvm_option(s?)
         | 'end'
@@ -425,30 +501,46 @@ $FAI::Parser = Parse::RecDescent->new(
         option(s?)
         | <error>
 
-    raid_option: /^preserve_always:(\d+(,\d+)*)/
+    raid_option: /^preserve_always:((\d+(,\d+)*)|all)/
         {
-          # set the preserve flag for all ids in all cases
-          $FAI::configs{RAID}{volumes}{$_}{preserve} = 1 foreach (split (",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{RAID}{opts_all}{preserve} = 1;
+          } else {
+            # set the preserve flag for all ids in all cases
+            $FAI::configs{RAID}{volumes}{$_}{preserve} = 1 foreach (split (",", $1));
+          }
         }
-        | /^preserve_reinstall:(\d+(,\d+)*)/
+        | /^preserve_reinstall:((\d+(,\d+)*)|all)/
         {
           # set the preserve flag for all ids if $FAI::reinstall is set
           if ($FAI::reinstall) {
-            $FAI::configs{RAID}{volumes}{$_}{preserve} = 1 foreach (split(",", $1));
+            if ($1 eq "all") {
+              $FAI::configs{RAID}{opts_all}{preserve} = 1;
+            } else {
+              $FAI::configs{RAID}{volumes}{$_}{preserve} = 1 foreach (split(",", $1));
+            }
           }
         }
-        | /^preserve_lazy:(\d+(,\d+)*)/
+        | /^preserve_lazy:((\d+(,\d+)*)|all)/
         {
-          $FAI::configs{RAID}{volumes}{$_}{preserve} = 2 foreach (split(",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{RAID}{opts_all}{preserve_lazy} = 1;
+          } else {
+            $FAI::configs{RAID}{volumes}{$_}{preserve} = 2 foreach (split(",", $1));
+          }
         }
         | /^fstabkey:(device|label|uuid)/
         {
           # the information preferred for fstab device identifieres
           $FAI::configs{$FAI::device}{fstabkey} = $1;
         }
-        | /^always_format:(\d+(,\d+)*)/
+        | /^always_format:((\d+(,\d+)*)|all)/
         {
-          $FAI::configs{RAID}{volumes}{$_}{always_format} = 1 foreach (split (",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{RAID}{opts_all}{always_format} = 1;
+          } else {
+            $FAI::configs{RAID}{volumes}{$_}{always_format} = 1 foreach (split (",", $1));
+          }
         }
 
     cryptsetup_option: /^randinit/
@@ -456,41 +548,57 @@ $FAI::Parser = Parse::RecDescent->new(
           $FAI::configs{$FAI::device}{randinit} = 1;
         }
 
-    lvm_option: m{^preserve_always:([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)}
+    lvm_option: m{^preserve_always:(([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)|all)}
         {
-          # set the preserve flag for all ids in all cases
-          foreach (split (",", $1)) {
-            (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or 
-              die &FAI::internal_error("VG re-parse failed");
-            $FAI::configs{"VG_$1"}{volumes}{$2}{size}{preserve} = 1;
-          }
-        }
-        | m{^preserve_reinstall:([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)}
-        {
-          # set the preserve flag for all ids if $FAI::reinstall is set
-          if ($FAI::reinstall) {
+          if ($1 eq "all") {
+            $FAI::configs{"VG_--ANY--"}{opts_all}{preserve} = 1;
+          } else {
+            # set the preserve flag for all ids in all cases
             foreach (split (",", $1)) {
-              (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or 
+              (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
                 die &FAI::internal_error("VG re-parse failed");
               $FAI::configs{"VG_$1"}{volumes}{$2}{size}{preserve} = 1;
             }
           }
         }
-        | m{^preserve_lazy:([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)}
+        | m{^preserve_reinstall:(([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)|all)}
         {
-          foreach (split (",", $1)) {
-            (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
-              die &FAI::internal_error("VG re-parse failed");
-            $FAI::configs{"VG_$1"}{volumes}{$2}{size}{preserve} = 2;
+          # set the preserve flag for all ids if $FAI::reinstall is set
+          if ($FAI::reinstall) {
+            if ($1 eq "all") {
+              $FAI::configs{"VG_--ANY--"}{opts_all}{preserve} = 1;
+            } else {
+              foreach (split (",", $1)) {
+                (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
+                  die &FAI::internal_error("VG re-parse failed");
+                $FAI::configs{"VG_$1"}{volumes}{$2}{size}{preserve} = 1;
+              }
+            }
           }
         }
-        | m{^resize:([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)}
+        | m{^preserve_lazy:(([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)|all)}
         {
-          # set the resize flag for all ids
-          foreach (split (",", $1)) {
-            (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or 
-              die &FAI::internal_error("VG re-parse failed");
-            $FAI::configs{"VG_$1"}{volumes}{$2}{size}{resize} = 1;
+          if ($1 eq "all") {
+            $FAI::configs{"VG_--ANY--"}{opts_all}{preserve_lazy} = 1;
+          } else {
+            foreach (split (",", $1)) {
+              (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
+                die &FAI::internal_error("VG re-parse failed");
+              $FAI::configs{"VG_$1"}{volumes}{$2}{size}{preserve} = 2;
+            }
+          }
+        }
+        | m{^resize:(([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)|all)}
+        {
+          if ($1 eq "all") {
+            $FAI::configs{"VG_--ANY--"}{opts_all}{resize} = 1;
+          } else {
+            # set the resize flag for all ids
+            foreach (split (",", $1)) {
+              (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
+                die &FAI::internal_error("VG re-parse failed");
+              $FAI::configs{"VG_$1"}{volumes}{$2}{size}{resize} = 1;
+            }
           }
         }
         | /^fstabkey:(device|label|uuid)/
@@ -498,47 +606,70 @@ $FAI::Parser = Parse::RecDescent->new(
           # the information preferred for fstab device identifieres
           $FAI::configs{"VG_--ANY--"}{fstabkey} = $1;
         }
-        | m{^always_format:([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)}
+        | m{^always_format:(([^/,\s\-]+-[^/,\s\-]+(,[^/,\s\-]+-[^/,\s\-]+)*)|all)}
         {
-          foreach (split (",", $1)) {
-            (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
-              die &FAI::internal_error("VG re-parse failed");
-            $FAI::configs{"VG_$1"}{volumes}{$2}{size}{always_format} = 1;
+          if ($1 eq "all") {
+            $FAI::configs{"VG_--ANY--"}{opts_all}{always_format} = 1;
+          } else {
+            foreach (split (",", $1)) {
+              (m{^([^/,\s\-]+)-([^/,\s\-]+)}) or
+                die &FAI::internal_error("VG re-parse failed");
+              $FAI::configs{"VG_$1"}{volumes}{$2}{size}{always_format} = 1;
+            }
           }
         }
 
 
-    option: /^preserve_always:(\d+(,\d+)*)/
+    option: /^preserve_always:((\d+(,\d+)*)|all)/
         {
-          # set the preserve flag for all ids in all cases
-          $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 1 foreach (split (",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{$FAI::device}{opts_all}{preserve} = 1;
+          } else {
+            # set the preserve flag for all ids in all cases
+            $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 1 foreach (split (",", $1));
+          }
           $FAI::configs{$FAI::device}{preserveparts} = 1;
         }
-        | /^preserve_reinstall:(\d+(,\d+)*)/
+        | /^preserve_reinstall:((\d+(,\d+)*)|all)/
         {
           # set the preserve flag for all ids if $FAI::reinstall is set
           if ($FAI::reinstall) {
-            $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 1 foreach (split(",", $1));
+            if ($1 eq "all") {
+              $FAI::configs{$FAI::device}{opts_all}{preserve} = 1;
+            } else {
+              $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 1 foreach (split(",", $1));
+            }
             $FAI::configs{$FAI::device}{preserveparts} = 1;
           }
         }
-        | /^preserve_lazy:(\d+(,\d+)*)/
+        | /^preserve_lazy:((\d+(,\d+)*)|all)/
         {
-          $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 2 foreach (split(",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{$FAI::device}{opts_all}{preserve_lazy} = 1;
+          } else {
+            $FAI::configs{$FAI::device}{partitions}{$_}{size}{preserve} = 2 foreach (split(",", $1));
+          }
           $FAI::configs{$FAI::device}{preserveparts} = 2;
         }
-        | /^resize:(\d+(,\d+)*)/
+        | /^resize:((\d+(,\d+)*)|all)/
         {
-          # set the resize flag for all ids
-          $FAI::configs{$FAI::device}{partitions}{$_}{size}{resize} = 1 foreach (split(",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{$FAI::device}{opts_all}{resize} = 1;
+          } else {
+            # set the resize flag for all ids
+            $FAI::configs{$FAI::device}{partitions}{$_}{size}{resize} = 1 foreach (split(",", $1));
+          }
           $FAI::configs{$FAI::device}{preserveparts} = 1;
         }
-        | /^disklabel:(msdos|gpt-bios|gpt)/
+        | /^disklabel:(\S+)/
         {
-          # set the disk label - actually not only the above, but all types 
+          my $dl = $1;
+          ($dl =~ /^(msdos|gpt-bios|gpt)$/) or die
+            "Invalid disk label $dl; use one of msdos|gpt-bios|gpt\n";
+          # set the disk label - actually not only the above, but all types
           # supported by parted could be allowed, but others are not implemented
           # yet
-          $FAI::configs{$FAI::device}{disklabel} = $1;
+          $FAI::configs{$FAI::device}{disklabel} = $dl;
         }
         | /^bootable:(\d+)/
         {
@@ -557,27 +688,37 @@ $FAI::Parser = Parse::RecDescent->new(
           # the information preferred for fstab device identifieres
           $FAI::configs{$FAI::device}{fstabkey} = $1;
         }
-	| /^sameas:disk(\d+)/
-	{
-	  my $ref_dev = &FAI::resolve_disk_shortname($1);
-	  defined($FAI::configs{"PHY_" . $ref_dev}) or die "Reference device $ref_dev not found in config\n";
-
-	  use Storable qw(dclone);
-
-	  $FAI::configs{$FAI::device} = dclone($FAI::configs{"PHY_" . $ref_dev});
-	}
 	| /^sameas:(\S+)/
 	{
 	  my $ref_dev = &FAI::resolve_disk_shortname($1);
 	  defined($FAI::configs{"PHY_" . $ref_dev}) or die "Reference device $ref_dev not found in config\n";
 
-	  use Storable qw(dclone);
-
-	  $FAI::configs{$FAI::device} = dclone($FAI::configs{"PHY_" . $ref_dev});
+	  $FAI::configs{$FAI::device} = Storable::dclone($FAI::configs{"PHY_" . $ref_dev});
+    # add entries to device tree
+    defined($FAI::dev_children{$ref_dev}) or
+      &FAI::internal_error("dev_children missing reference entry");
+    ($FAI::device =~ /^PHY_(.+)$/) or
+      &FAI::internal_error("unexpected device name");
+    my $disk = $1;
+    foreach my $p (@{ $FAI::dev_children{$ref_dev} }) {
+      my ($i_p_d, $rd, $pd) = &FAI::phys_dev($p);
+      (1 == $i_p_d) or next;
+      ($rd eq $ref_dev) or &FAI::internal_error("dev_children is inconsistent");
+      push @{ $FAI::dev_children{$disk} }, &FAI::make_device_name($disk, $pd);
+    }
 	}
-        | /^always_format:(\d+(,\d+)*)/
+        | /^always_format:((\d+(,\d+)*)|all)/
         {
-          $FAI::configs{$FAI::device}{partitions}{$_}{size}{always_format} = 1 foreach (split(",", $1));
+          if ($1 eq "all") {
+            $FAI::configs{$FAI::device}{opts_all}{always_format} = 1;
+          } else {
+            $FAI::configs{$FAI::device}{partitions}{$_}{size}{always_format} = 1 foreach (split(",", $1));
+          }
+        }
+        | /^align-at:(\d+)([kKMGTPiB]+)?/
+        {
+          my $u = defined($2) ? $2 : "MiB";
+          $FAI::configs{$FAI::device}{align_at} = &FAI::convert_unit("$1$u") * 1024.0 * 1024.0;
         }
 
     volume: /^vg\s+/ name devices vgcreateopt(s?)
@@ -598,18 +739,21 @@ $FAI::Parser = Parse::RecDescent->new(
           $FAI::configs{RAID}{volumes}{$vol_id}{mode} = $1;
           # initialise the hash of devices
           $FAI::configs{RAID}{volumes}{$vol_id}{devices} = {};
-          # initialise the preserve flag
+          # initialise the flags
           defined($FAI::configs{RAID}{volumes}{$vol_id}{preserve}) or
-            $FAI::configs{RAID}{volumes}{$vol_id}{preserve} = 0;
-          # initialise the always_format flag
+            $FAI::configs{RAID}{volumes}{$vol_id}{preserve} =
+              defined($FAI::configs{RAID}{opts_all}{preserve}) ? 1 :
+                (defined($FAI::configs{RAID}{opts_all}{preserve_lazy}) ? 2 : 0);
           defined($FAI::configs{RAID}{volumes}{$vol_id}{always_format}) or
-            $FAI::configs{RAID}{volumes}{$vol_id}{always_format} = 0;
+            $FAI::configs{RAID}{volumes}{$vol_id}{always_format} =
+              defined($FAI::configs{RAID}{opts_all}{always_format}) ? 1 : 0;
           # set the reference to the current volume
           # the reference is used by all further processing of this config line
           $FAI::partition_pointer = (\%FAI::configs)->{RAID}->{volumes}->{$vol_id};
+          $FAI::partition_pointer_dev_name = "/dev/md$vol_id";
         }
         mountpoint devices filesystem mount_options mdcreateopts
-        | /^(luks|tmp|swap)\s+/
+        | /^(luks|luks:"[^"]+"|tmp|swap)\s+/
         {
           ($FAI::device eq "CRYPT") or
             die "Encrypted device spec $1 invalid in context $FAI::device\n";
@@ -627,6 +771,7 @@ $FAI::Parser = Parse::RecDescent->new(
           $FAI::configs{CRYPT}{volumes}{$vol_id}{preserve} = 0;
 
           $FAI::partition_pointer = (\%FAI::configs)->{CRYPT}->{volumes}->{$vol_id};
+          $FAI::partition_pointer_dev_name = "CRYPT$vol_id";
         }
         mountpoint devices filesystem mount_options lv_or_fsopts
         | /^tmpfs\s+/
@@ -647,6 +792,7 @@ $FAI::Parser = Parse::RecDescent->new(
           $FAI::configs{TMPFS}{volumes}{$vol_id}{preserve} = 0;
 
           $FAI::partition_pointer = (\%FAI::configs)->{TMPFS}->{volumes}->{$vol_id};
+          $FAI::partition_pointer_dev_name = "TMPFS$vol_id";
         }
         mountpoint tmpfs_size mount_options
         | type mountpoint size filesystem mount_options lv_or_fsopts
@@ -662,6 +808,12 @@ $FAI::Parser = Parse::RecDescent->new(
           # initialise a logical partition
           &FAI::init_part_config($item[ 1 ]);
         }
+        | 'raw-disk'
+        {
+          # initialise a pseudo-partition: this disk will be used without
+          # partitioning it
+          &FAI::init_part_config("raw");
+        }
         | m{^([^/,\s\-]+)-([^/,\s\-]+)\s+}
         {
           # set $FAI::device to VG_$1
@@ -670,21 +822,30 @@ $FAI::Parser = Parse::RecDescent->new(
           defined ($FAI::configs{$FAI::device}{devices}) or
             die "Volume group $1 has not been declared yet.\n";
           # make sure, $2 has not been defined already
-          defined ($FAI::configs{$FAI::device}{volumes}{$2}{size}{range}) and 
+          defined ($FAI::configs{$FAI::device}{volumes}{$2}{size}{range}) and
             die "Logical volume $2 has been defined already.\n";
+          # add to ordered list
+          push @{ $FAI::configs{$FAI::device}{ordered_lv_list} }, $2;
           # initialise the new hash
           defined($FAI::configs{$FAI::device}{volumes}{$2}) or
             $FAI::configs{$FAI::device}{volumes}{$2} = {};
-          # initialise the preserve and resize flags
+          # initialise the flags
           defined($FAI::configs{$FAI::device}{volumes}{$2}{size}{preserve}) or
-            $FAI::configs{$FAI::device}{volumes}{$2}{size}{preserve} = 0;
+            $FAI::configs{$FAI::device}{volumes}{$2}{size}{preserve} =
+              defined($FAI::configs{"VG_--ANY--"}{opts_all}{preserve}) ? 1 :
+                (defined($FAI::configs{"VG_--ANY--"}{opts_all}{preserve_lazy}) ? 2 : 0);
           defined($FAI::configs{$FAI::device}{volumes}{$2}{size}{always_format}) or
-            $FAI::configs{$FAI::device}{volumes}{$2}{size}{always_format} = 0;
+            $FAI::configs{$FAI::device}{volumes}{$2}{size}{always_format} =
+              defined($FAI::configs{"VG_--ANY--"}{opts_all}{always_format}) ? 1 : 0;
           defined($FAI::configs{$FAI::device}{volumes}{$2}{size}{resize}) or
-            $FAI::configs{$FAI::device}{volumes}{$2}{size}{resize} = 0;
+            $FAI::configs{$FAI::device}{volumes}{$2}{size}{resize} =
+              defined($FAI::configs{"VG_--ANY--"}{opts_all}{resize}) ? 1 : 0;
           # set the reference to the current volume
           # the reference is used by all further processing of this config line
           $FAI::partition_pointer = (\%FAI::configs)->{$FAI::device}->{volumes}->{$2};
+          $FAI::partition_pointer_dev_name = "/dev/$1/$2";
+          # add entry to device tree
+          push @{ $FAI::dev_children{$FAI::device} }, $FAI::partition_pointer_dev_name;
         }
 
     mountpoint: m{^(-|swap|/[^\s\:]*)(:encrypt(:randinit)?)?}
@@ -718,11 +879,15 @@ $FAI::Parser = Parse::RecDescent->new(
             $FAI::configs{$FAI::device}{volumes} = {};
           # initialise the list of physical devices
           $FAI::configs{$FAI::device}{devices} = ();
+          # initialise the ordered list of volumes
+          $FAI::configs{$FAI::device}{ordered_lv_list} = ();
+          # init device tree
+          $FAI::dev_children{$FAI::device} = ();
           # the rule must not return undef
           1;
         }
 
-    size: /^((RAM:\d+%|\d+[kMGTP%iB]*)(-(RAM:\d+%|\d+[kMGTP%iB]*)?)?)(:resize)?/
+    size: /^((RAM:\d+%|\d+[kKMGTP%iB]*)(-(RAM:\d+%|\d+[kKMGTP%iB]*)?)?)(:resize|:preserve_(always|reinstall|lazy))?/
         {
           # complete the size specification to be a range in all cases
           my $range = $1;
@@ -736,45 +901,18 @@ $FAI::Parser = Parse::RecDescent->new(
           {
             # range has no upper limit, assume the whole disk
             $range = "${range}100%";
-          } 
+          }
 
-          # convert the units, if necessary
-          my ($min, $max) = split (/-/, $range);
-          $min .= "MiB" if ($min =~ /\d\s*$/);
-          $min   = &FAI::convert_unit($min);
-          $max .= "MiB" if ($max =~ /\d\s*$/);
-          $max   = &FAI::convert_unit($max);
-          $range = "$min-$max";
-          # enter the range into the hash
-          $FAI::partition_pointer->{size}->{range} = $range;
-          # set the resize flag, if required
-          if (defined ($5)) {
-            $FAI::partition_pointer->{size}->{resize} = 1;
-            $FAI::configs{$FAI::device}{preserveparts} = 1;
-          }
+          &FAI::set_volume_size($range, $5);
         }
-        | /^(-(RAM:\d+%|\d+[kMGTP%iB]*))(:resize)?\s+/
+        | /^(-(RAM:\d+%|\d+[kKMGTP%iB]*))(:resize|:preserve_(always|reinstall|lazy))?\s+/
         {
-          # complete the range by assuming 0 as the lower limit 
-          my $range = "0$1";
-          # convert the units, if necessary
-          my ($min, $max) = split (/-/, $range);
-          $min .= "MiB" if ($min =~ /\d\s*$/);
-          $min   = &FAI::convert_unit($min);
-          $max .= "MiB" if ($max =~ /\d\s*$/);
-          $max   = &FAI::convert_unit($max);
-          $range = "$min-$max";
-          # enter the range into the hash
-          $FAI::partition_pointer->{size}->{range} = $range;
-          # set the resize flag, if required
-          if (defined ($3)) {
-            $FAI::partition_pointer->{size}->{resize} = 1;
-            $FAI::configs{$FAI::device}{preserveparts} = 1;
-          }
+          # complete the range by assuming 0 as the lower limit
+          &FAI::set_volume_size("0$1", $3);
         }
         | <error: invalid partition size near "$text">
 
-    tmpfs_size: /^(RAM:(\d+%)|\d+[kMGTPiB]*)\s+/
+    tmpfs_size: /^(RAM:(\d+%)|\d+[kKMGTPiB]*)\s+/
         {
           my $size;
 
@@ -801,7 +939,7 @@ $FAI::Parser = Parse::RecDescent->new(
           foreach my $dev (split(",", $1))
           {
             # match the substrings
-            ($dev =~ /^([^\d,:\s\-][^,:\s]*)(:(spare|missing))*$/) or 
+            ($dev =~ /^([^\d,:\s\-][^,:\s]*)(:(spare|missing))*$/) or
               &FAI::internal_error("PARSER ERROR");
             # redefine the device string
             $dev = $1;
@@ -840,22 +978,28 @@ $FAI::Parser = Parse::RecDescent->new(
                 $dev = $candidates[0];
               }
               # each device may only appear once
-              defined ($FAI::partition_pointer->{devices}->{$dev}) and 
+              defined ($FAI::partition_pointer->{devices}->{$dev}) and
                 die "$dev is already part of the RAID volume\n";
               # set the options
               $FAI::partition_pointer->{devices}->{$dev} = {
                 "spare" => $spare,
                 "missing" => $missing
               };
+              # add entry to device tree
+              push @{ $FAI::dev_children{$dev} }, $FAI::partition_pointer_dev_name;
             } elsif ($FAI::device eq "CRYPT") {
               die "Failed to resolve $dev to a unique device name\n" if (scalar(@candidates) != 1);
               $FAI::partition_pointer->{device} = $candidates[0];
               &FAI::mark_encrypted($candidates[0]);
+              # add entry to device tree
+              push @{ $FAI::dev_children{$candidates[0]} }, $FAI::partition_pointer_dev_name;
             } else {
               die "Failed to resolve $dev to a unique device name\n" if (scalar(@candidates) != 1);
               $dev = $candidates[0];
               # create an empty hash for each device
               $FAI::configs{$FAI::device}{devices}{$dev} = {};
+              # add entry to device tree
+              push @{ $FAI::dev_children{$dev} }, $FAI::device;
             }
           }
           1;
@@ -965,5 +1109,85 @@ sub run_parser {
   defined $FAI::Parser->file($input) or die "Syntax error\n";
 }
 
+################################################################################
+#
+# @brief Check for invalid configs (despite correct syntax)
+#
+################################################################################
+sub check_config {
+
+  my %all_mount_pts = ();
+
+  # loop through all configs
+  foreach my $config (keys %FAI::configs) {
+    if ($config =~ /^PHY_(.+)$/) {
+      (scalar(keys %{ $FAI::configs{$config}{partitions} }) > 0) or
+        die "Empty disk_config stanza for device $1\n";
+      foreach my $p (keys %{ $FAI::configs{$config}{partitions} }) {
+        next if (1 == $FAI::configs{$config}{partitions}{$p}{size}{extended});
+        defined($FAI::configs{$config}{partitions}{$p}{mountpoint}) or
+          &FAI::internal_error("Undefined mountpoint for non-extended partition");
+        my $this_mp = $FAI::configs{$config}{partitions}{$p}{mountpoint};
+        next if ($this_mp eq "-");
+        defined($all_mount_pts{$this_mp}) and die
+          "Mount point $this_mp used twice\n";
+        defined($FAI::dev_children{&FAI::make_device_name($1, $p)}) and die
+          "Mount point $this_mp is shadowed by stacked devices\n";
+        ($this_mp eq "none") or $all_mount_pts{$this_mp} = 1;
+      }
+    } elsif ($config =~ /^VG_(.+)$/) {
+      next if ($1 eq "--ANY--");
+      (scalar(keys %{ $FAI::configs{$config}{volumes} }) ==
+        scalar(@{ $FAI::configs{$config}{ordered_lv_list} })) or
+        &FAI::internal_error("Inconsistent LV lists - missing entries");
+      defined($FAI::configs{$config}{volumes}{$_}) or
+        &FAI::internal_error("Inconsistent LV lists - missing entries")
+        foreach (@{ $FAI::configs{$config}{ordered_lv_list} });
+      foreach my $p (keys %{ $FAI::configs{$config}{volumes} }) {
+        my $this_mp = $FAI::configs{$config}{volumes}{$p}{mountpoint};
+        next if ($this_mp eq "-");
+        defined($all_mount_pts{$this_mp}) and die
+          "Mount point $this_mp used twice\n";
+        defined($FAI::dev_children{"/dev/$1/$p"}) and die
+          "Mount point $this_mp is shadowed by stacked devices\n";
+        ($this_mp eq "none") or $all_mount_pts{$this_mp} = 1;
+      }
+      next;
+    } elsif ($config eq "RAID") {
+      (scalar(keys %{ $FAI::configs{$config}{volumes} }) > 0) or
+        die "Empty RAID configuration\n";
+      foreach my $p (keys %{ $FAI::configs{$config}{volumes} }) {
+        my $this_mp = $FAI::configs{$config}{volumes}{$p}{mountpoint};
+        next if ($this_mp eq "-");
+        defined($all_mount_pts{$this_mp}) and die
+          "Mount point $this_mp used twice\n";
+        defined($FAI::dev_children{"/dev/md$p"}) and die
+          "Mount point $this_mp is shadowed by stacked devices\n";
+        ($this_mp eq "none") or $all_mount_pts{$this_mp} = 1;
+      }
+    } elsif ($config eq "CRYPT") {
+      foreach my $p (keys %{ $FAI::configs{$config}{volumes} }) {
+        my $this_mp = $FAI::configs{$config}{volumes}{$p}{mountpoint};
+        next if ($this_mp eq "-");
+        defined($all_mount_pts{$this_mp}) and die
+          "Mount point $this_mp used twice\n";
+        ($this_mp eq "none") or $all_mount_pts{$this_mp} = 1;
+      }
+      next;
+    } elsif ($config eq "TMPFS") {
+      foreach my $p (keys %{ $FAI::configs{$config}{volumes} }) {
+        my $this_mp = $FAI::configs{$config}{volumes}{$p}{mountpoint};
+        next if ($this_mp eq "-");
+        defined($all_mount_pts{$this_mp}) and die
+          "Mount point $this_mp used twice\n";
+        ($this_mp eq "none") or $all_mount_pts{$this_mp} = 1;
+      }
+      next;
+    } else {
+      &FAI::internal_error("Unexpected key $config");
+    }
+  }
+}
+
 1;
 
